@@ -16,7 +16,7 @@ import textwrap
 import time
 from urllib.request import urlopen
 
-__version__ = "0.95"
+__version__ = "0.98"
 
 DOCKER_NS = "satex"
 REGISTRY_URL = "https://github.com/sat-heritage/docker-images/releases/download/list/list.tgz"
@@ -42,13 +42,6 @@ IN_REPOSITORY = in_repository()
 ##
 #
 # List of images
-def make_name(reg, cfg, entry, solver):
-    pattern = cfg[entry].get("image_name", "{SOLVER}:{ENTRY}")
-    return pattern.format(ENTRY=entry, SOLVER=solver)
-
-def images_of_registry(reg, cfg):
-    return [make_name(reg, cfg, entry, solver) \
-                for entry in reg for solver in reg[entry]]
 
 def fetch_registry(args, opener):
     with opener("index.json") as fp:
@@ -92,17 +85,38 @@ def get_registry(args):
             refresh_cache(args, force=True)
             return get_registry(args)
 
+def make_name(reg, cfg, entry, solver):
+    pattern = cfg[entry].get("image_name", "{SOLVER}:{ENTRY}")
+    return pattern.format(ENTRY=entry, SOLVER=solver)
+
 class Repository(object):
     def __init__(self, args):
         self.registry, self.setup = get_registry(args)
         self.images = {}
+        self.names = []
+
+        select_all = not hasattr(args, "all") or args.all
+        select_fixme = select_all or hasattr(args, "fixme") and args.fixme
+        select_unstable = select_all or hasattr(args, "unstable") and args.unstable
+        select_stable = select_all or (not select_unstable and not select_fixme)
+
         for entry in self.registry:
             for solver in self.registry[entry]:
                 name = make_name(self.registry, self.setup, entry, solver)
                 if hasattr(args, "pattern") and \
                         not fnmatch.fnmatch(name, args.pattern):
                     continue
+                status = self.registry[entry][solver].get("status", "unknown")
+                if status == "ok":
+                    if not select_stable:
+                        continue
+                elif status.startswith("FIXME"):
+                    if not select_fixme:
+                        continue
+                elif not select_unstable:
+                    continue
                 self.images[name] = {"entry": entry, "solver": solver}
+                self.names.append(name)
 
 class ImageManager(object):
     def __init__(self, name, repo):
@@ -117,13 +131,13 @@ class ImageManager(object):
     @property
     def solver_name(self):
         return self.registry["name"]
+    @property
+    def status(self):
+        return self.registry.get("status", "unknown")
 
 
 def get_list(args):
-    reg, cfg = get_registry(args)
-    images = images_of_registry(reg, cfg)
-    if hasattr(args, "pattern"):
-        images = fnmatch.filter(images, args.pattern)
+    images = Repository(args).names
     assert len(images), "No matching images!"
     return images
 
@@ -132,45 +146,63 @@ def print_list(args):
         print(image)
 
 
-_info_first = ["name", "version", "authors"]
+_info_first = ["name", "version", "authors", "base_from"]
+_info_last = ["comment", "comments"]
 _info_ignore = {"call"}
 def print_info(args):
     repo = Repository(args)
     for name in repo.images:
         image = ImageManager(name, repo)
 
-        keys = [k for k in sorted(image.registry.keys()) if k not in _info_ignore]
-        for k in _info_first[::-1]:
-            if k in keys:
-                keys.remove(k)
-                keys.insert(0, k)
+        ignore = _info_ignore.union(_info_first).union(_info_last)
+        keys = _info_first \
+            + [k for k in sorted(image.registry.keys()) if k not in ignore] \
+            + _info_last
 
         key_width = 0
         info = []
         for key in keys:
-            value = image.registry[key]
-            if key == "args":
-                key = "Call"
-                value = f"{image.registry['call']} {' '.join(value)}"
-            elif key == "argsproof":
-                key = "Call (proof)"
-                value = f"{image.registry['call']} {' '.join(value)}"
+            if key == "base_from":
+                name = "Environment"
+                value = image.setup[key]
+                builder_base = image.setup.get("builder_base")
+                if builder_base:
+                    if builder_base == value:
+                        value += " (same as build)"
+                    else:
+                        value += f" (build: {builder_base})"
+            else:
+                if key not in image.registry:
+                    continue
+                name = key.title()
+                value = image.registry[key]
+            if key in ["args", "argsproof"]:
+                name = "Call"
+                if key == "argsproof":
+                    name += " (proof)"
+                value = f"{image.registry['call']} {' '.join(map(str,value))}"
             else:
-                key = key.title()
                 value = str(value)
-            key_width = max(len(key), key_width)
-            info.append((key,value))
+            key_width = max(len(name), key_width)
+            info.append((name,value))
 
         key_width += 2
         line_width = key_width + 70
-        print(f"{DOCKER_NS}/\033[1m{name}\033[0m")
+        if image.status == "ok":
+            color = 32
+        elif image.status.startswith("FIXME"):
+            color = 31
+        else:
+            color = 33
+        print(f"{DOCKER_NS}/\033[1;{color}m{image.name}\033[0m")
         print("-"*line_width)
         for (key, value) in info:
             key = f"{key}: "
-            for line in textwrap.wrap(value):
-                print("{0:{key_width}}{1}".format(key, line,
-                    key_width=key_width))
-                key = ""
+            for p in value.splitlines():
+                for line in textwrap.wrap(p):
+                    print("{0:{key_width}}{1}".format(key, line,
+                        key_width=key_width))
+                    key = ""
         print()
 
 #
@@ -230,10 +262,11 @@ def prepare_image(args, docker_argv, image):
         info(" ".join(cmd))
         subprocess.check_call(cmd)
 
+_docker_opts = []
 def docker_runs(args, images, docker_args=(), image_args=()):
     docker_argv = check_docker()
     argv = ["run", "--rm"]
-    for opt in ["volume"]:
+    for opt in _docker_opts:
         if getattr(args, opt) is not None:
             val = getattr(args, opt)
             if isinstance(val, list):
@@ -254,7 +287,7 @@ def docker_runs(args, images, docker_args=(), image_args=()):
         else:
             info(" ".join(cmd))
             ret = subprocess.call(cmd)
-            assert ret == 0 or 10 <= ret <= 20, "Solver failed!"
+            assert ret == 0 or 10 <= ret <= 20, f"Solver failed! ({ret})"
     if not args.pretend:
         return ret
 
@@ -321,6 +354,8 @@ def docker_build(args, docker_argv, tag, root, build_args={}, Dockerfile=None):
         for f in FROMs:
             docker_uptodate_image(args, docker_argv, f)
     argv = docker_argv + ["build", "-t", tag, root]
+    if args.no_cache:
+        argv += ["--no-cache"]
     if Dockerfile:
         argv += ["-f", Dockerfile]
     for k,v in build_args.items():
@@ -334,6 +369,15 @@ def build_images(args):
 
     bases_uptodate = set()
 
+    only_dist_opts = ["RDEPENDS"]
+    hide_opts = [
+        "base_version",
+        "base_from",
+        "builder",
+        "builder_base",
+        "image_name",
+    ]
+
     for name in repo.images:
         image = ImageManager(name, repo)
         setup = image.setup
@@ -345,15 +389,9 @@ def build_images(args):
 
         root = str(image.entry)
 
-        build_args = {k:v for k,v in setup.items() if \
-                        k not in ["generic_version",
-                                    "builder", "builder_base",
-                                    "image_name"] and isinstance(v, str)}
-        build_args["solver"] = image.solver_name
-        build_args["solver_id"] = image.solver
-        for k in setup:
-            if k in build_args:
-                build_args[k] = build_args[k].format(**fmtvars)
+        build_args = {k: v.format(**fmtvars) for k,v in setup.items() if \
+                k not in hide_opts and isinstance(v, str)}
+        build_args.update(fmtvars)
 
         builder_path = setup["builder"]
         if not builder_path.startswith("generic/"):
@@ -368,20 +406,32 @@ def build_images(args):
         docker_build(args, docker_argv, builder_target, root,
                 build_args=build_args, Dockerfile=builder_Dockerfile)
 
-        base = f"base:{setup['base_version']}"
-        if base not in bases_uptodate:
-            docker_build(args, docker_argv, f"{DOCKER_NS}/{base}", base.replace(":", "/"))
-            bases_uptodate.add(base)
+        base_version = setup["base_version"]
+        base_root = os.path.join("base", base_version)
+        base_from = setup.get("base_from")
+        base_args = {}
+        base_tag = base_version
+        if base_from:
+            base_args["BASE"] = base_from
+            from_tag = base_from.replace("/","_").replace(":","-")
+            base_tag = f"{base_version}-{from_tag}"
+        base_target = f"{DOCKER_NS}/base:{base_tag}"
+        if base_target not in bases_uptodate:
+            docker_build(args, docker_argv, base_target, base_root, base_args)
+            bases_uptodate.add(base_target)
 
         dist_version = setup.get("dist_version", "v1")
         dist_Dockerfile = f"generic/dist-{dist_version}/Dockerfile"
         dist_args = {
-            "BASE": f"{DOCKER_NS}/{base}",
+            "BASE": base_target,
             "BUILDER_BASE": builder_target,
             "IMAGE_NAME": image.name,
-            "solver": build_args["solver"],
-            "solver_id": build_args["solver_id"],
+            "SOLVER": build_args["SOLVER"],
+            "SOLVER_NAME": build_args["SOLVER_NAME"],
         }
+        for k in only_dist_opts:
+            if k in setup:
+                dist_args[k] = setup[k]
 
         fd, dbjson = tempfile.mkstemp(".json", "file", root)
         try:
@@ -399,10 +449,10 @@ def build_images(args):
 def test_images(args):
     images = get_list(args)
     docker_args = ["-v", f"{os.path.abspath('tests')}:/data"]
-    for dimacs in glob.glob("tests/*.gz"):
-        info(f"Testing {dimacs}")
-        dimacs = os.path.basename(dimacs)
-        docker_runs(args, images, docker_args, image_args=(dimacs,))
+    dimacs = args.cnf
+    info(f"Testing {dimacs}")
+    dimacs = os.path.basename(dimacs)
+    docker_runs(args, images, docker_args, image_args=(dimacs,))
 
 def push_images(args):
     docker_argv = check_docker()
@@ -462,15 +512,29 @@ def main(redirected=False):
     ##
     # options shared by several sub-commands
     #
-    spec_parser = argparse.ArgumentParser(add_help=False)
+    status_parser = argparse.ArgumentParser(add_help=False)
+    status_parser.add_argument("--unstable", action="store_true",
+            help="Consider images with non-ok and non-FIXME status")
+    status_parser.add_argument("--fixme", action="store_true",
+            help="Consider images with FIXME status")
+    status_parser.add_argument("--all", "-a", action="store_true",
+            help="Consider all images, with any status")
+
+    spec_parser = argparse.ArgumentParser(add_help=False,
+            parents=[status_parser])
     spec_parser.add_argument("pattern",
             help="Pattern for filtering images")
 
+
     docker_parser = argparse.ArgumentParser(add_help=False)
     docker_parser.add_argument("--pull", action="store_true",
             help="Explicitly pull the image")
     docker_parser.add_argument("-v", "--volume", action="append",
             help="(Docker option) Mount a volume")
+    _docker_opts.append("volume")
+    docker_parser.add_argument("-e", "--env", action="append",
+            help="(Docker option) Set environment variables")
+    _docker_opts.append("env")
     #
     ##
 
@@ -479,7 +543,8 @@ def main(redirected=False):
     #
 
     p = subparsers.add_parser("list",
-            help=f"List {DOCKER_NS} Docker images")
+            help=f"List {DOCKER_NS} Docker images",
+            parents=[status_parser])
     p.add_argument("pattern", default="*", nargs="?",
             help="Pattern for filtering images (default: *)")
     p.set_defaults(func=print_list)
@@ -527,11 +592,14 @@ def main(redirected=False):
         p = subparsers.add_parser("build",
                 help=f"Build {DOCKER_NS} Docker images",
                 parents=[spec_parser])
+        p.add_argument("--no-cache", action="store_true",
+                help="docker build option")
         p.set_defaults(func=build_images)
 
         p = subparsers.add_parser("test",
                 help=f"Test {DOCKER_NS} Docker images",
                 parents=[spec_parser, docker_parser])
+        p.add_argument("--cnf", "-f", default="tests/cmu-bmc-barrel6.cnf.gz")
         p.set_defaults(func=test_images)
 
         p = subparsers.add_parser("push",
